#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module classes
#
# Copyright (C) 2017-2020 Univention GmbH <http://www.univention.de/>
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

from ldap.filter import filter_format
import os
import json
import listener
import univention.uldap
import univention.admin.uldap
import univention.admin.objects
from univention.ox.oxlogger import OXLogger
from univention.ox.listener_tools import AsRoot, get_credentials_for_context, ListenerTools_RuntimeError, ox_call, ox_list_group


class ContinueHookCommand(object):
	pass


class ExitHookCommand(object):
	def __init__(self, exit_code=None):
		self.exit_code = exit_code


class OxListener(object):
	CSV_GROUP_ROW_GROUPNAME = "name"
	CSV_GROUP_ROW_MEMBERS = "members"
	CSV_GROUP_ROW_ID = 'id'
	CSV_USER_ROW_DEFAULT_SENDER_ADDRESS = "DefaultSenderAddress"
	CSV_USER_ROW_DISPLAYNAME = 'displayname'
	CSV_USER_ROW_GROUPNAME = 'name'
	CSV_USER_ROW_MEMBERS = 'members'
	CSV_USER_ROW_USERNAME = 'Name'
	CSV_USER_ROW_UID = 'Id'

	_ldap_machine_connection = None
	_ldap_admin_connection = None
	_udm_modules_ro = dict()
	_udm_modules_rw = dict()

	def __init__(self, name, old, new, command=None):  # type: (str, Dict[str, List[str]], Dict[str, List[str]], Optional[str]) -> None
		self.module_name = name
		self.old = old
		self.new = new
		self.command = command
		self.logger = OXLogger(self.module_name)

	@classmethod
	def get_ldap_machine_connection(cls):  # type: () -> Tuple[object, object]
		if not cls._ldap_machine_connection:
			with AsRoot():
				try:
					cls._ldap_machine_connection = univention.admin.uldap.getMachineConnection(ldap_master=False)
				except IOError as ex:
					raise ListenerTools_RuntimeError(ex)
		return cls._ldap_machine_connection

	@classmethod
	def get_ldap_admin_connection(cls):  # type: () -> Tuple[object, object]
		if not cls._ldap_admin_connection:
			with AsRoot():
				try:
					cls._ldap_admin_connection = univention.admin.uldap.getAdminConnection()
				except IOError as ex:
					raise ListenerTools_RuntimeError(ex)
		return cls._ldap_admin_connection

	@classmethod
	def get_credentials(cls, context_id):  # type: (str) -> Tuple[str, str]
		return get_credentials_for_context(context_id)

	@classmethod
	def get_context_ids_from_group(cls, old_or_new):  # type: (Dict[str, List[str]]) -> List[str]
		context_ids = old_or_new.get('oxContextIDNum')
		if context_ids:
			return context_ids
		else:
			context_objs = cls.get_udm_ox_contexts()
			return [c['contextid'] for c in context_objs]

	@classmethod
	def get_udm_group(cls, groupdn, modify=False):  # type: (str, Optional[bool]) -> object
		return cls.get_udm_obj('groups/group', groupdn, modify)

	@classmethod
	def get_udm_ox_groups(cls, filter_s=None, modify=False):  # type: (str, Optional[str], Optional[bool]) -> List[object]
		return cls.find_udm_objects(
			'groups/group',
			'(&(isOxGroup=OK){})'.format(filter_s if filter_s else ''),
			listener.configRegistry["ldap/base"],
			modify)

	@classmethod
	def get_udm_ox_contexts(cls, modify=False):  # type: (str, Optional[bool]) -> List[object]
		return cls.find_udm_objects('oxmail/oxcontext', 'cn=*', listener.configRegistry["ldap/base"], modify)

	@classmethod
	def find_udm_objects(cls, module_name, filter_s, base, modify=False):  # type: (str, str, str, Optional[bool]) -> List[object]
		"""
		search LDAP for UDM objects
		:param module_name: str: "users/user", "groups/group", etc
		:param filter_s: str: LDAP filter string
		:param base: str: node to start search from
		:param modify: bool: whether an admin connection should be used
		:return: list of (not yet opened) UDM objects
		"""
		lo, po, mod = cls.init_udm(module_name, modify)
		return mod.lookup(None, lo, filter_s=filter_s, base=base)

	@classmethod
	def get_udm_user(cls, userdn, attributes=None, modify=False):  # type: (str, Optional[List[str]], Optional[bool]) -> object
		return cls.get_udm_obj("users/user", userdn, attributes, modify)

	@classmethod
	def get_udm_obj(cls, module_name, dn, attributes=None, modify=False):  # type: (str, str, str, Optional[List[str]], Optional[bool]) -> object
		lo, po, mod = cls.init_udm(module_name, modify)
		obj = mod.object(None, lo, po, dn, attributes=attributes)
		obj.open()
		return obj

	@classmethod
	def init_udm(cls, module_name, modify=False):  # type: (str, Optional[bool]) -> Tuple(object, object, object)
		if modify:
			lo, po = cls.get_ldap_admin_connection()
			mod_cache = cls._udm_modules_rw
		else:
			lo, po = cls.get_ldap_machine_connection()
			mod_cache = cls._udm_modules_ro
		try:
			mod = mod_cache[module_name]
		except KeyError:
			univention.admin.modules.update()
			mod = univention.admin.modules.get(module_name)
			with AsRoot():
				univention.admin.modules.init(lo, po, mod)
				mod_cache[module_name] = mod
		return lo, po, mod

	@classmethod
	def get_ldap_group_displayname(cls, group_name):  # type: (str) -> str
		lo, po = cls.get_ldap_machine_connection()
		filter_s = filter_format('(&(cn=%s)(objectClass=oxGroup))', (group_name,))
		result_id = lo.search(base=listener.configRegistry['ldap/base'], filter=filter_s, attr=['description'])
		try:
			return result_id[0][1]['description'][0]
		except (TypeError, KeyError, IndexError):
			return group_name

	@classmethod
	def get_ox_groups(cls, context_id):  # type: (str) -> Dict[str, List[str]]
		admuser, admpass = cls.get_credentials(context_id)
		res = dict()
		for row in ox_list_group(admuser, admpass, context_id):
			res[row[cls.CSV_GROUP_ROW_GROUPNAME]] = row[cls.CSV_GROUP_ROW_MEMBERS].split(',')
		return res

	def remove_group(self, groupname, context_id):  # type: (str, str) -> None
		self.logger.process('Deleting OX group "{}" in context {}...'.format(groupname, context_id))
		admuser, admpass = self.get_credentials(context_id)
		delete_command = [
			'deletegroup',
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(groupname),
		]
		ox_call(delete_command, 'com.openexchange.admin.rmi.exceptions.NoSuchObjectException', self.logger)

	def remove_user(self, username, context_id):  # type: (str, str) -> None
		self.logger.process('Deleting OX user "{}" in context {}...'.format(username, context_id))
		cmd = ["deleteuser", "--username={}".format(username)]
		cmd.extend(self._auth_args(context_id))
		ox_call(cmd, "com.openexchange.admin.rmi.exceptions.NoSuchObjectException", self.logger)

	def _auth_args(self, context_id):  # type: (str) -> List[str]
		admuser, admpass = self.get_credentials(context_id)
		return [
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
		]

	@staticmethod
	def modrdn_load_old(old, path):
		if os.path.exists(path):
			with open(path, 'rb') as fp:
				old = json.load(fp)
			os.unlink(path)
		return old

	@staticmethod
	def modrdn_save_old(old, path):
		with open(path, "wb+") as fp:
			os.chmod(path, 0o600)
			json.dump(old, fp)

	def run(self):  # type: () -> None
		raise NotImplementedError
